
// Operations on stepped block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRangeRep': stepped block-distributed template range.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* StepRangeRep :: subrng(const int _extent,
                                 const int _base, const int _stride) {
  return new StepSubRangeRep(_extent, _base, _stride, this,
                              kernel, blockSize, stepCrd) ;
}

Format StepRangeRep :: format() const {
  return DIST_STEP ;
}

int StepRangeRep :: ker_glb(const int glb, int* sub) const {

  if(glb < stepSub) {
    *sub = glb % (blockSize + 1) ;
    return glb / (blockSize + 1) ;
  }
  else {
    *sub = (glb - stepCrd) % blockSize ;
    return (glb - stepCrd) / blockSize ;
  }
}

void StepRangeRep :: location(Location* loc, const int glb) const {

  if(glb < stepSub) {
    loc->subs [0] = glb % (blockSize + 1) ;
    loc->crd      = glb / (blockSize + 1) ;
  }
  else {
    loc->subs [0] = (glb - stepCrd) % blockSize ;
    loc->crd      = (glb - stepCrd) / blockSize ;
  }
  loc->dim     = dimension ;
}

RangeRep* StepRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

int StepRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  // `ker_glb' is a global subscript in the kernel range.

  *sub = 0 ;
  if(ker_glb < stepCrd) {
    *count = blockSize + 1 ;
    return (blockSize + 1) * ker_glb ;
  }
  else {
    *count = blockSize ;
    return blockSize * ker_glb + stepCrd ;
  }
}

int StepRangeRep :: block_step(int* sub_stp) const {
  *sub_stp = 1 ;

  return 1 ;
}

int StepRangeRep :: width() const {
  return stepCrd ? blockSize + 1 : blockSize ;
}

int StepRangeRep :: glb(const int sub, const int ker_glb) const {
  if(ker_glb < stepCrd)
    return (blockSize + 1) * ker_glb + sub ;
  else
    return blockSize * ker_glb + stepCrd + sub ;
}

void StepRangeRep :: show() {
  AdlibPrintf("StepRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(StepRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepSubRangeRep': stepped block-distributed subrange.        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* StepSubRangeRep :: subrng(const int _extent,
                                    const int _base, const int _stride) {
  return new StepSubRangeRep(_extent,
			     base + stride * _base, stride * _stride,
			     parent, kernel, blockSize, stepCrd) ;
}

RangeRep* StepSubRangeRep :: tem() {return parent ; }

int StepSubRangeRep :: str() const {return stride ; }

int StepSubRangeRep :: bas() const {return base ; }

Format StepSubRangeRep :: format() const {
  return DIST_STEP ;
}

int StepSubRangeRep :: ker_glb(const int glb, int* sub) const {
  const int tem = base + stride * glb ;

  if(tem < stepSub) {
    *sub = tem % (blockSize + 1) ;
    return tem / (blockSize + 1) ;
  }
  else {
    *sub = (tem - stepCrd) % blockSize ;
    return (tem - stepCrd) / blockSize ;
  }
}

void StepSubRangeRep :: location(Location* loc, const int glb) const {
  const int tem = base + stride * glb ;

  if(glb < stepSub) {
    loc->subs [0] = tem % (blockSize + 1) ;
    loc->crd      = tem / (blockSize + 1) ;
  }
  else {
    loc->subs [0] = (tem - stepCrd) % blockSize ;
    loc->crd      = (tem - stepCrd) / blockSize ;
  }
  loc->dim     = dimension ;
}

RangeRep* StepSubRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

int StepSubRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  if(ker_glb >= ker_lo && ker_glb <= ker_hi) {
    int blockStart, blockStartOffset, blockFinishOffset ;
    if(ker_glb < stepCrd) {
      blockStart        = ker_glb * (blockSize + 1) ;
      blockStartOffset  = blockStart - base ;
      blockFinishOffset = blockStartOffset + blockSize ;
    }
    else {
      blockStart        = ker_glb * blockSize + stepCrd;
      blockStartOffset  = blockStart - base ;
      blockFinishOffset = blockStartOffset + blockSize - 1 ;
    }

    int i_lo, i_hi ;

    if(stride > 0) {
      i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    }
    else {
      i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    }

    *sub   = base + stride * i_lo - blockStart ;
    *count = i_hi - i_lo + 1 ;

    return i_lo ;
  }
  else {
    *sub   = 0 ;
    *count = 0 ;

    return 0 ;
  }
}

int StepSubRangeRep :: block_step(int* sub_stp) const {
  *sub_stp = stride ;

  return 1 ;
}

RangeRep* StepSubRangeRep :: subker(int* ker_bas, int* ker_str) const {

  if(stride > 0) {
    *ker_bas = ker_lo ;
    *ker_str = 1 ;
  }
  else {
    *ker_bas = ker_hi ;
    *ker_str = -1 ;
  }

  return kernel->subrng(ker_hi - ker_lo + 1, *ker_bas, *ker_str) ;
}

int StepSubRangeRep :: width() const {
  return stepCrd ? blockSize + 1 : blockSize ;
}

int StepSubRangeRep :: glb(const int sub, const int ker_glb) const {
  if(ker_glb < stepCrd)
    return ((blockSize + 1) * ker_glb + sub - base) / stride ;
  else
    return (blockSize * ker_glb + stepCrd + sub - base) / stride ;
}

void StepSubRangeRep :: show() {
  AdlibPrintf("StepSubRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(StepSubRangeRep)


// -------------------------------------------------------------
// Things to do
//
//   `subker' returns non-empty subrange even for extent = 0
//        (not a problem?)

